#
# Copyright (c) SAS Institute Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import itertools
import logging
import re

from lxml import etree
import six

from .artifact import Artifact
from .utils import memoize
from .utils import parse_source
from .versioning import VersionRange

EMPTY_POM = """\
<?xml version="1.0" encoding="UTF-8"?>
<project xmlns="http://maven.apache.org/POM/4.0.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
         xsi:schemaLocation="http://maven.apache.org/POM/4.0.0 http://maven.apache.org/xsd/maven-4.0.0.xsd">
          <modelVersion>4.0.0</modelVersion>
          <groupId>{0.group_id}</groupId>
          <artifactId>{0.artifact_id}</artifactId>
          <version>{0.version}</version>
</project>
"""
POM_NAMESPACE = "http://maven.apache.org/POM/4.0.0"
POM = "{%s}" % (POM_NAMESPACE,)
POM_NAMESPACE_LEN = len(POM)
POM_PARSER = etree.XMLParser(
    recover=True,
    remove_comments=True,
    remove_pis=True,
    )
PROPERTY_RE = re.compile(r'\$\{(.*?)\}')
STRIP_NAMESPACE_RE = re.compile(POM)

log = logging.getLogger(__name__)


class Pom(Artifact):
    """Parse a pom file into a python object
    """

    RANGE_CHARS = ('[', '(', ']', ')')

    __slots__ = ("_client", "_parent", "_dep_mgmt", "_dependencies", "_pom_data", "_properties")

    def __init__(self, coordinate, client=None, pom_data=None):
        if pom_data is not None:
            pom_data = etree.fromstring(pom_data.encode("utf-8"), parser=POM_PARSER)
        self._pom_data = pom_data
        self._client = client

        # dynamic attributes
        self._parent = None
        self._dep_mgmt = None
        self._dependencies = None
        self._properties = None
        super(Pom, self).__init__(coordinate)

    def _find_deps(self, elem=None):
        if elem is None:
            elem = self.pom_data
        dependencies = {}

        # find all non-optional, compile dependencies
        deps = _find(elem, "dependencies")
        if deps is None:
            return dependencies
        for elem in _findall(deps, "dependency"):
            group = _findtext(elem, "groupId")
            if group is None:
                log.warning("dependency element without groupId: '%s'", etree.tostring(elem))
                continue
            group = self._replace_properties(group)
            artifact = _findtext(elem, "artifactId")
            if artifact is None:
                log.warning("dependency element without artifactId: '%s'", etree.tostring(elem))
                continue
            artifact = self._replace_properties(artifact)

            if (group, artifact) in self.dependency_management:
                version, scope, optional = self.dependency_management[(group, artifact)]
            else:
                version = optional = scope = None
            local_version = _findtext(elem, "version")
            if local_version is not None:
                version = local_version
            elif version is None:
                # FIXME: Default to the latest released version if no
                # version is specified. I'm not sure if this is the
                # correct behavior, but let's try it for now.
                version = "latest.release"
            version = self._replace_properties(version)

            local_scope = _findtext(elem, "scope")
            if local_scope is not None:
                scope = local_scope
            elif scope is None:
                # if scope is None, then it should be "compile"
                scope = "compile"
            scope = scope.strip()

            local_optional = _findtext(elem, "optional")
            if local_optional is not None:
                optional = local_optional.strip()
            elif optional is None:
                # default optional to false
                optional = "false"
            # convert optional to boolean
            optional = optional == "true"
            dependencies.setdefault(scope, set()).add(((group, artifact, version), not optional))
        return dependencies

    def _find_dependency_management(self, elem=None):
        if elem is None:
            elem = self.pom_data
        dep_mgmt = {}
        import_mgmt = {}

        dependency_management = _find(elem, "dependencyManagement")
        if dependency_management is None:
            return import_mgmt
        dependencies = _find(dependency_management, "dependencies")
        if dependencies is None:
            return import_mgmt
        for elem in _findall(dependencies, "dependency"):
            group = _findtext(elem, "groupId")
            if group is None:
                log.warning("dependencyManagement/dependencies/dependency element without groupId: '%s'",
                            etree.tostring(elem))
                continue
            group = self._replace_properties(group)
            artifact = _findtext(elem, "artifactId")
            if artifact is None:
                log.warning("dependencyManagement/dependencies/dependency element without artifactId: '%s'",
                            etree.tostring(elem))
                continue
            artifact = self._replace_properties(artifact)
            version = _findtext(elem, "version")
            if version is None:
                log.warning("dependencyManagement/dependencies/dependency element without version: '%s'",
                            etree.tostring(elem))
                continue
            version = self._replace_properties(version)

            optional = _findtext(elem, "optional")
            optional = (optional is not None and optional == "true")
            scope = _findtext(elem, "scope")
            if scope is not None:
                scope = scope.strip()
            if scope == "import":
                import_pom = self._pom_factory(group, artifact, version)
                import_mgmt.update(import_pom.dependency_management)
            dep_mgmt[(group, artifact)] = (version, scope, optional)

        import_mgmt.update(dep_mgmt)
        return import_mgmt

    def _find_import_deps(self):
        dependencies = {}
        # process dependency management to find imports
        for group, artifact in self.dependency_management:
            version, scope, optional = self.dependency_management[(group, artifact)]
            if scope == "import":
                dependencies.setdefault(scope, set()).add(((group, artifact, version), not optional))
        return dependencies

    def _find_prerequisites(self, elem=None):
        if elem is None:
            elem = self.pom_data
        properties = {}
        # get prerequisites
        prereqs = _find(elem, "prerequisites")
        if prereqs is None:
            return properties
        for elem in prereqs:
            tag = elem.tag[POM_NAMESPACE_LEN:]
            properties['prerequisites.' + tag] = elem.text
            properties['project.prerequisites.' + tag] = elem.text
        return properties

    def _find_profiles(self, elem=None):
        if elem is None:
            elem = self.pom_data
        active_profiles = []
        default_profiles = []
        profiles = _find(elem, "profiles")
        if profiles is None:
            return default_profiles
        for p in _findall(profiles, "profile"):
            by_default = _findtext(p, "activation/activeByDefault")
            by_default = by_default is not None and by_default == "true"
            if by_default:
                default_profiles.append(p)
            else:
                jdk = _findtext(p, "activation/jdk")
                if jdk is not None:
                    # attempt some clean up
                    if (jdk.startswith('[') or jdk.startswith("![")) \
                            and jdk.endswith(','):
                        # assume they left off the )
                        jdk += ')'

                    # TODO: make the JDK version selectable
                    if jdk.startswith('!'):
                        vr = VersionRange.fromstring(jdk[1:])
                        if (vr.version and "1.8" != vr.version) \
                                or (not vr.version and "1.8" not in vr):
                            active_profiles.append(p)
                    else:
                        vr = VersionRange.fromstring(jdk)
                        if (vr.version and "1.8" == vr.version) \
                                or (not vr.version and "1.8" in vr):
                            active_profiles.append(p)
        if active_profiles:
            return active_profiles
        return default_profiles

    def _find_properties(self, elem=None):
        if elem is None:
            elem = self.pom_data
        properties = {}
        project_properties = _find(elem, "properties")
        if project_properties is not None:
            for prop in project_properties.iterchildren():
                if prop.tag == POM + 'property':
                    name = prop.get('name')
                    value = prop.get('value')
                else:
                    name = prop.tag[POM_NAMESPACE_LEN:]
                    value = prop.text
                properties[name] = value
        return properties

    def _find_relocations(self, elem=None):
        if elem is None:
            elem = self.pom_data
        dependencies = {}
        # process distributionManagement for relocation
        distManagement = _find(elem, "distributionManagement")
        if distManagement is None:
            return dependencies
        relocation = _find(distManagement, "relocation")
        if relocation is None:
            return dependencies
        group = _findtext(relocation, "groupId")
        if group is None:
            group = self.group_id
        else:
            group = self._replace_properties(group)

        artifact = _findtext(relocation, "artifactId")
        if artifact is None:
            artifact = self.artifact_id
        else:
            artifact = self._replace_properties(artifact)

        version = _findtext(relocation, "version")
        if version is None:
            version = self.version
        else:
            version = self._replace_properties(version)
        dependencies.setdefault("relocation", set()).add(((group, artifact, version), True))
        return dependencies

    def _pom_factory(self, group, artifact, version):
        return Pom("%s:%s:pom:%s" % (group, artifact, version), self._client)

    def _replace_properties(self, text, properties=None):
        if properties is None:
            properties = self.properties

        def subfunc(matchobj):
            key = matchobj.group(1)
            return properties.get(key)

        result = PROPERTY_RE.sub(subfunc, text)
        while result and PROPERTY_RE.match(result):
            result = PROPERTY_RE.sub(subfunc, result)

        if not result:
            result = text
        return result.strip()

    def pick_version(self, spec, artifacts):
        """Pick a version from *versions* according to the spec

        Convert spec into maven version range and return the first version in
        *versions* that is within the range.

        :param str spec: a maven version range spec or gradle dynamic version
        :param versions: list of available versions for this artifact
        :type versions: [:py:class:`pymaven.Version`, ...]
        :return: the newest version that matches the spec
        :rtype: str or None
        """
        if spec in ("latest.release", "release"):
            for a in artifacts:
                if 'snapshot' not in str(a.version.version).lower():
                    return str(a.version)
        elif spec in ("latest.integration", "latest"):
            return str(artifacts[0].version)

        range = VersionRange.fromstring(spec)
        for artifact in artifacts:
            if artifact.version in range:
                return str(artifact.version)

    @property
    @memoize("_dependencies")
    def dependencies(self):
        dependencies = {}
        # we depend on our parent
        if self.parent is not None:
            group = self.parent.group_id
            artifact = self.parent.artifact_id
            version = self.parent.version
            dependencies.setdefault("compile", set()).add(((group, artifact, str(version)), True))
        for key, value in itertools.chain(
                six.iteritems(self._find_import_deps()),
                six.iteritems(self._find_deps()),
                six.iteritems(self._find_relocations())):
            dependencies.setdefault(key, set()).update(value)
        for profile in self._find_profiles():
            for key, value in itertools.chain(
                    six.iteritems(self._find_deps(profile)),
                    six.iteritems(self._find_relocations(profile))):
                dependencies.setdefault(key, set()).update(value)
        return dependencies

    @property
    @memoize("_dep_mgmt")
    def dependency_management(self):
        dep_mgmt = {}
        # add parent's block first so we can override it
        if self.parent is not None:
            dep_mgmt.update(self.parent.dependency_management)
        dep_mgmt.update(self._find_dependency_management())
        for profile in self._find_profiles():
            dep_mgmt.update(self._find_dependency_management(profile))
        return dep_mgmt

    @property
    @memoize("_parent")
    def parent(self):
        parent = _find(self.pom_data, "parent")
        if parent is not None:
            group = _findtext(parent, "groupId").strip()
            artifact = _findtext(parent, "artifactId").strip()
            version = _findtext(parent, "version").strip()
            return self._pom_factory(group, artifact, version)

    @property
    @memoize("_pom_data")
    def pom_data(self):
        if self._client is None:
            return etree.fromstring(EMPTY_POM.format(self), parser=POM_PARSER)
        with self._client.get_artifact(self.coordinate).contents as fh:
            return etree.parse(fh, parser=POM_PARSER)

    @property
    @memoize("_properties")
    def properties(self):
        properties = {}
        if self.parent is not None:
            properties.update(self.parent.properties)
            properties['parent.groupId'] = self.parent.group_id
            properties['parent.artifactId'] = self.parent.artifact_id
            properties['parent.version'] = str(self.parent.version)
            properties['project.parent.groupId'] = self.parent.group_id
            properties['project.parent.artifactId'] = self.parent.artifact_id
            properties['project.parent.version'] = str(self.parent.version)
        # built-in properties
        properties['artifactId'] = self.artifact_id
        properties['groupId'] = self.group_id
        properties['version'] = str(self.version)
        properties['project.artifactId'] = self.artifact_id
        properties['project.groupId'] = self.group_id
        properties['project.version'] = str(self.version)
        properties['pom.artifactId'] = self.artifact_id
        properties['pom.groupId'] = self.group_id
        properties['pom.version'] = str(self.version)
        properties.update(self._find_properties())
        properties.update(self._find_prerequisites())
        for profile in self._find_profiles():
            properties.update(self._find_properties(profile))
        return properties

    def get_dependencies(self):
        return set(self.iter_dependencies())

    def get_build_dependencies(self):
        return set(self.iter_build_dependencies())

    def iter_dependencies(self):
        return itertools.chain(*self.dependencies.values())

    def iter_build_dependencies(self):
        return itertools.chain(
            (d for d, r in self.dependencies.get("compile", set()) if r),
            (d for d, r in self.dependencies.get("import", set()) if r),
            (d for d, r in self.dependencies.get("relocation", set()) if r),
            )

    @classmethod
    def parse(cls, coordinate, source, client=None):
        """Return a :ref:`Pom` object loaded with source. ``source`` can be any
        of the following:

        - a file name/path
        - a file-like object
        - a URL with the file, http(s), or ftp scheme

        :param str coordinate: the maven coordinates of the POM
        :param source: source to parse
        :param client: a :class:`MavenClient`
        :returns: a :ref:`Pom` object
        """
        fh = parse_source(source)
        return cls(coordinate, pom_data=fh.read(), client=client)

    @classmethod
    def fromstring(cls, coordinate, text, client=None):
        """Parses a POM document from a string.

        :param str coordinate: the maven coordinates of the POM
        :param str text: text to parse
        :param client: a :class:`MavenClient`
        :returns: a :ref:`Pom` object
        """
        return cls(coordinate, pom_data=text, client=client)


def _find(elem, tag):
    tag = tag.replace("/", "/" + POM)
    return elem.find(POM + tag)


def _findall(elem, tag):
    tag = tag.replace("/", "/" + POM)
    return elem.findall(POM + tag)


def _findtext(elem, tag):
    tag = tag.replace("/", "/" + POM)
    return elem.findtext(POM + tag)



#
# Copyright (c) SAS Institute Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import unittest

from pymaven import Artifact
from pymaven.errors import ArtifactParseError


class TestArtifact(unittest.TestCase):
    def _assertArtifactOrder(self, a1, a2):
        assert a1 < a2
        assert a2 > a1
        assert not a1 >= a2
        assert not a2 <= a1
        assert not a1 == a2
        assert a1 != a2

    def test_invalid_coordinate(self):
        tests = ("foo", "1:2:3:4:5:6")
        for input in tests:
            self.assertRaises(ArtifactParseError, Artifact, input)

    def test_path(self):
        test_pairs = (
            ("foo.bar:baz", "foo/bar/baz"),
            ("foo.bar:baz:1", "foo/bar/baz/1/baz-1.jar"),
            ("foo.bar:baz:pkg:1", "foo/bar/baz/1/baz-1.pkg"),
            ("foo.bar:baz:pkg:sources:1", "foo/bar/baz/1/baz-1-sources.pkg"),
            ("foo.bar:baz:pkg:javadoc:1", "foo/bar/baz/1/baz-1-javadoc.pkg"),
            ("foo.bar:baz:[1,2)", "foo/bar/baz"),
            )

        for input, expected in test_pairs:
            artifact = Artifact(input)
            assert artifact.path == expected

    def test_comparison(self):
        test_pairs = (
            # compare group id
            (Artifact("f:a:1"), Artifact("g:a:1")),
            # compare artifact id
            (Artifact("g:a:1"), Artifact("g:b:1")),
            # compare type
            (Artifact("g:a:1"), Artifact("g:a:pom:1")),
            (Artifact("g:a:jar:1"), Artifact("g:a:pom:1")),
            (Artifact("g:a:pom:1"), Artifact("g:a:war:1")),
            # compare classifier
            (Artifact("g:a:jar:c:1"), Artifact("g:a:1")),
            (Artifact("g:a:jar:a:1"), Artifact("g:a:jar:c:1")),
            # compare version
            (Artifact("g:a:1"), Artifact("g:a:2")),
            # mask version
            (Artifact("f:a:2"), Artifact("g:a:1")),
            (Artifact("g:a:2"), Artifact("g:b:1")),
            (Artifact("g:a:2"), Artifact("g:a:pom:1")),
            (Artifact("g:a:jar:2"), Artifact("g:a:pom:1")),
            (Artifact("g:a:pom:2"), Artifact("g:a:war:1")),
            (Artifact("g:a:jar:c:2"), Artifact("g:a:1")),
            (Artifact("g:a:jar:a:2"), Artifact("g:a:jar:c:1")),
            )

        for pair in test_pairs:
            self._assertArtifactOrder(*pair)

        # verify identity
        a = Artifact("foo:bar:1")
        assert not a < a
        assert a <= a
        assert a >= a
        assert a == a
        assert not a != a

        # compare to non-artifact
        assert a > "aardvark"
        assert a != "aardvark"
        assert "aardvark" < a
        assert "aardvark" != a
        assert a > 10
        assert a != 10
        assert 10 < a
        assert 10 != a

    def test_tostring(self):
        a = Artifact("foo:bar")
        assert str(a) == "foo:bar"

        a = Artifact("foo:bar:1")
        assert str(a) == "foo:bar:jar:1"

        a = Artifact("foo:bar:pom:1")
        assert str(a) == "foo:bar:pom:1"

        a = Artifact("foo:bar:pom:sources:1")
        assert str(a) == "foo:bar:pom:sources:1"



#
# Copyright (c) SAS Institute Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""
Test version handling
"""
import unittest

from pymaven import Version
from pymaven import VersionRange
from pymaven.errors import RestrictionParseError
from pymaven.errors import VersionRangeParseError
from pymaven.versioning import Restriction


class TestRestriction(unittest.TestCase):
    def test_everyting_spec(self):
        r = Restriction("[,)")
        assert r.lower_bound is None
        assert r.lower_bound_inclusive
        assert r.upper_bound is None
        assert not r.upper_bound_inclusive

        assert "0.1.0" in r
        assert "1.0" in r
        assert "2.0" in r

    def test_exclusive_lower_bound(self):
        r = Restriction("(1.0,2.0]")
        assert str(r.lower_bound) == "1.0"
        assert not r.lower_bound_inclusive
        assert str(r.upper_bound) == "2.0"
        assert r.upper_bound_inclusive

        assert "0.1" not in r
        assert "1.0" not in r
        assert "1.1" in r

    def test_no_lower_limit(self):
        r = Restriction("(,1.0]")
        assert r.lower_bound is None
        assert not r.lower_bound_inclusive
        assert str(r.upper_bound) == "1.0"
        assert r.upper_bound_inclusive
        assert "0.1" in r
        assert "1.0" in r
        assert "1.1" not in r
        assert "2.0" not in r

    def test_inclusive_range(self):
        r = Restriction("[1.0]")
        assert str(r.lower_bound) == "1.0"
        assert r.lower_bound_inclusive
        assert str(r.upper_bound) == "1.0"
        assert r.upper_bound_inclusive

        assert "0.8" not in r
        assert "1.0" in r
        assert "1.1" not in r
        assert "2.0" not in r

        r = Restriction("[1.2,1.3]")
        assert str(r.lower_bound) == "1.2"
        assert r.lower_bound_inclusive
        assert str(r.upper_bound) == "1.3"
        assert r.upper_bound_inclusive

        assert "0.8" not in r
        assert "1.1" not in r
        assert "1.2" in r
        assert "1.2.1" in r
        assert "1.3" in r
        assert "1.3.1" not in r
        assert "2.0" not in r

    def test_exclusive_upper_bound(self):
        r = Restriction("[1.0,2.0)")
        assert str(r.lower_bound) == "1.0"
        assert r.lower_bound_inclusive
        assert str(r.upper_bound) == "2.0"
        assert not r.upper_bound_inclusive

        assert "0.8" not in r
        assert "1.0" in r
        assert "1.9" in r
        assert "2.0" not in r
        assert "3.0" not in r

        r = Restriction("[1.5,)")
        assert str(r.lower_bound) == "1.5"
        assert r.lower_bound_inclusive
        assert r.upper_bound is None
        assert not r.upper_bound_inclusive

        assert "0.8" not in r
        assert "1.4" not in r
        assert "1.5" in r
        assert "100.3" in r

    def test_invalid_restrictions(self):
        tests = ("(1.0)", "[1.0)", "(1.0]", "(1.0,1.0]", "[1.0,1.0)",
                 "(1.0,1.0)", "[1.1,1.0]",)
        for spec in tests:
            self.assertRaises(
                RestrictionParseError,
                Restriction,
                spec,
                )

    def test_compare(self):
        r1 = Restriction("(1.0,2.0]")

        assert r1 == r1
        assert r1 == "(1.0,2.0]"
        assert 1 < r1

    def test_string_repr(self):
        for input in ("[1.0]", "[1.0,)", "[1.0,2.0]", "[1.0,2.0)", "(1.0,2.0)",
                      "[,2.0]", "[,2.0)", "(,2.0)"):
            actual = str(Restriction(input))
            assert input == actual, \
                "Restriction(%s) == %s, wanted %s" % (input, actual, input)


class TestVersion(unittest.TestCase):
    """Tests the Version object"""
    def _assert_version_equal(self, v1, v2):
        V1 = Version(v1)
        V2 = Version(v2)
        # test Version equal Version
        assert V1 == V2, \
            "%s != %s" % (V1, V2)
        assert V2 == V1, \
            "%s != %s" % (V2, V1)

        # test str equal Version
        assert v1 == V2, \
            "%s != %s" % (v1, V2)
        assert v2 == V1, \
            "%s != %s" % (v2, V1)

        # test Version equal str
        assert V1 == v2, \
            "%s != %s" % (V1, v2)
        assert V2 == v1, \
            "%s != %s" % (V2, v1)

    def _assert_version_order(self, v1, v2):
        V1 = Version(v1)
        V2 = Version(v2)

        # Version <-> Version order
        assert V1 < V2, \
            "%s >= %s" % (V1._parsed, V2._parsed)
        assert V2 > V1, \
            "%s >= %s" % (V2._parsed, V1._parsed)

        # Version <-> str order
        assert V1 < v2, \
            "%s >= %s" % (V1._parsed, v2)
        assert V2 > v1, \
            "%s >= %s" % (V2._parsed, v1)

        # str <-> Version order
        assert v1 < V2, \
            "%s >= %s" % (v1, V2._parsed)
        assert v2 > V1, \
            "%s >= %s" % (v2, V1._parsed)

    def test_from_string(self):
        test_pairs = (
            # weird versions
            (".1", (0, 1)), ("-1", ((1,),)),
            # test some major.minor.tiny parsing
            ("1", (1,)), ("1.0", (1,)), ("1.0.0", (1,)),
            ("1.0.0.0", (1,)), ("11", (11,)), ("11.0", (11,)),
            ("1-1", (1, (1,))), ("1-1-1", (1, (1, (1,)))), (" 1 ", (1,)),
            # test qualifeirs
            ("1.0-ALPHA", (1, ("alpha",))), ("1-alpha", (1, ("alpha",))),
            ("1.0ALPHA", (1, ("alpha",))), ("1-alpha", (1, ("alpha",))),
            ("1.0-A", (1, ("a",))), ("1-a", (1, ("a",))),
            ("1.0A", (1, ("a",))), ("1a", (1, ("a",))),
            ("1.0-BETA", (1, ("beta",))), ("1-beta", (1, ("beta",))),
            ("1.0-B", (1, ("b",))), ("1-b", (1, ("b",))),
            ("1.0B", (1, ("b",))), ("1b", (1, ("b",))),
            ("1.0-MILESTONE", (1, ("milestone",))),
            ("1.0-milestone", (1, ("milestone",))), ("1-M", (1, ("m",))),
            ("1.0-m", (1, ("m",))), ("1M", (1, ("m",))), ("1m", (1, ("m",))),
            ("1.0-RC", (1, ("rc",))), ("1-rc", (1, ("rc",))),
            ("1.0-SNAPSHOT", (1, ("snapshot",))),
            ("1.0-snapshot", (1, ("snapshot",))), ("1-SP", (1, ("sp",))),
            ("1.0-sp", (1, ("sp",))), ("1-GA", (1,)), ("1-ga", (1,)),
            ("1.0-FINAL", (1,)), ("1-final", (1,)),
            ("1.0-CR", (1, ("rc",))), ("1-cr", (1, ("rc",))),
            # test some transistion
            ("1.0-alpha1", (1, ("alpha", (1,)))),
            ("1.0-alpha2", (1, ("alpha", (2,)))),
            ("1.0.0alpha1", (1, ("alpha", (1,)))),
            ("1.0-beta1", (1, ("beta", (1,)))),
            ("1-beta2", (1, ("beta", (2,)))),
            ("1.0.0beta1", (1, ("beta", (1,)))),
            ("1.0-BETA1", (1, ("beta", (1,)))),
            ("1-BETA2", (1, ("beta", (2,)))),
            ("1.0.0BETA1", (1, ("beta", (1,)))),
            ("1.0-milestone1", (1, ("milestone", (1,)))),
            ("1.0-milestone2", (1, ("milestone", (2,)))),
            ("1.0.0milestone1", (1, ("milestone", (1,)))),
            ("1.0-MILESTONE1", (1, ("milestone", (1,)))),
            ("1.0-milestone2", (1, ("milestone", (2,)))),
            ("1.0.0MILESTONE1", (1, ("milestone", (1,)))),
            ("1.0-alpha2snapshot", (1, ("alpha", (2, ("snapshot",))))),
            )

        for test, expected in test_pairs:
            v = Version(test)
            assert v._parsed == expected, \
                "Version(%s) == %s, want %s" % (test, v._parsed, expected)

    def test_version_qualifiers(self):
        version_qualifiers = (
            "1-alpha2snapshot", "1-alpha2", "1-alpha-123", "1-beta-2",
            "1-beta123", "1-m2", "1-m11", "1-rc", "1-cr2", "1-rc123",
            "1-SNAPSHOT", "1", "1-sp", "1-sp2", "1-sp123", "1-abc", "1-def",
            "1-pom-1", "1-1-snapshot", "1-1", "1-2", "1-123",
            )
        for idx, low in enumerate(version_qualifiers[:-1]):
            for high in version_qualifiers[idx+1:]:
                self._assert_version_order(low, high)

    def test_version_numbers(self):
        version_numbers = (
            "2.0", "2-1", "2.0.a", "2.0.0.a", "2.0.2", "2.0.123", "2.1.0",
            "2.1-a", "2.1b", "2.1-x", "2.1-1", "2.1.0.1", "2.2", "2.123",
            "11.a2", "11.a11", "11.b2", "11.b11", "11.m2", "11.m11", "11",
            "11.a", "11b", "11c", "11m",
            )
        for idx, low in enumerate(version_numbers[:-1]):
            for high in version_numbers[idx+1:]:
                self._assert_version_order(low, high)

        unicode_version_numbers = (
            # again, but with unicode input
            u"2.0", u"2-1", u"2.0.a", u"2.0.0.a", u"2.0.2", u"2.0.123",
            u"2.1.0", u"2.1-a", u"2.1b", u"2.1-x", u"2.1-1", u"2.1.0.1", u"2.2",
            u"2.123", u"11.a2", u"11.a11", u"11.b2", u"11.b11", u"11.m2",
            u"11.m11", u"11", u"11.a", u"11b", u"11c", u"11m",
            )
        for idx, low in enumerate(unicode_version_numbers[:-1]):
            for high in unicode_version_numbers[idx+1:]:
                self._assert_version_order(low, high)

    def test_version_equality(self):
        self._assert_version_equal("1", "1")
        self._assert_version_equal("1", "1.0")
        self._assert_version_equal("1", "1.0.0")
        self._assert_version_equal("1.0", "1.0.0")
        self._assert_version_equal("1", "1-0")
        self._assert_version_equal("1", "1.0-0")
        self._assert_version_equal("1.0", "1.0-0")
        # no separator between number and character
        self._assert_version_equal("1a", "1-a")
        self._assert_version_equal("1a", "1.0-a")
        self._assert_version_equal("1a", "1.0.0-a")
        self._assert_version_equal("1.0a", "1-a")
        self._assert_version_equal("1.0.0a", "1-a")
        self._assert_version_equal("1x", "1-x")
        self._assert_version_equal("1x", "1.0-x")
        self._assert_version_equal("1x", "1.0.0-x")
        self._assert_version_equal("1.0x", "1-x")
        self._assert_version_equal("1.0.0x", "1-x")

        # aliases
        self._assert_version_equal("1ga", "1")
        self._assert_version_equal("1final", "1")
        self._assert_version_equal("1cr", "1rc")

        # special "aliases" a, b and m for alpha, beta and milestone
        self._assert_version_equal("1a1", "1-alpha-1")
        self._assert_version_equal("1b2", "1-beta-2")
        self._assert_version_equal("1m3", "1-milestone-3")

        # case insensitive
        self._assert_version_equal("1X", "1x")
        self._assert_version_equal("1A", "1a")
        self._assert_version_equal("1B", "1b")
        self._assert_version_equal("1M", "1m")
        self._assert_version_equal("1Ga", "1")
        self._assert_version_equal("1GA", "1")
        self._assert_version_equal("1Final", "1")
        self._assert_version_equal("1FinaL", "1")
        self._assert_version_equal("1FINAL", "1")
        self._assert_version_equal("1Cr", "1Rc")
        self._assert_version_equal("1cR", "1rC")
        self._assert_version_equal("1m3", "1Milestone3")
        self._assert_version_equal("1m3", "1MileStone3")
        self._assert_version_equal("1m3", "1MILESTONE3")

        # unicode
        self._assert_version_equal(u"1", "1")
        self._assert_version_equal(u"1", "1.0")
        self._assert_version_equal(u"1", "1.0.0")
        self._assert_version_equal(u"1.0", "1.0.0")
        self._assert_version_equal(u"1", "1-0")
        self._assert_version_equal(u"1", "1.0-0")
        self._assert_version_equal(u"1.0", "1.0-0")

        self._assert_version_equal("1", u"1")
        self._assert_version_equal("1", u"1.0")
        self._assert_version_equal("1", u"1.0.0")
        self._assert_version_equal("1.0", u"1.0.0")
        self._assert_version_equal("1", u"1-0")
        self._assert_version_equal("1", u"1.0-0")
        self._assert_version_equal("1.0", u"1.0-0")

        self._assert_version_equal(u"1", u"1")
        self._assert_version_equal(u"1", u"1.0")
        self._assert_version_equal(u"1", u"1.0.0")
        self._assert_version_equal(u"1.0", u"1.0.0")
        self._assert_version_equal(u"1", u"1-0")
        self._assert_version_equal(u"1", u"1.0-0")
        self._assert_version_equal(u"1.0", u"1.0-0")

    def test_version_compare(self):
        self._assert_version_order("1", "2")
        self._assert_version_order("1.5", "2")
        self._assert_version_order("1", "2.5")
        self._assert_version_order("1.0", "1.1")
        self._assert_version_order("1.1", "1.2")
        self._assert_version_order("1.0.0", "1.1")
        self._assert_version_order("1.0.1", "1.1")
        self._assert_version_order("1.1", "1.2.0")
        self._assert_version_order("1.0-alpha-1", "1.0")
        self._assert_version_order("1.0-alpha-1", "1.0-alpha-2")
        self._assert_version_order("1.0-alpha-1", "1.0-beta-1")
        self._assert_version_order("1.0-beta-1", "1.0-SNAPSHOT")
        self._assert_version_order("1.0-SNAPSHOT", "1.0")
        self._assert_version_order("1.0-alpha-1-SNAPSHOT", "1.0-alpha-1")
        self._assert_version_order("1.0", "1.0-1")
        self._assert_version_order("1.0-1", "1.0-2")
        self._assert_version_order("1.0.0", "1.0-1")
        self._assert_version_order("2.0-1", "2.0.1")
        self._assert_version_order("2.0.1-klm", "2.0.1-lmn")
        self._assert_version_order("2.0.1", "2.0.1-xyz")
        self._assert_version_order("2.0.1", "2.0.1-123")
        self._assert_version_order("2.0.1-xyz", "2.0.1-123")
        # unicode input
        self._assert_version_order(u"1", "2")
        self._assert_version_order(u"1.5", "2")
        self._assert_version_order(u"1", "2.5")
        self._assert_version_order(u"1.0", "1.1")
        self._assert_version_order(u"1.1", "1.2")
        self._assert_version_order(u"1.0.0", "1.1")
        self._assert_version_order(u"1.0.1", "1.1")
        self._assert_version_order(u"1.1", "1.2.0")
        self._assert_version_order(u"1.0-alpha-1", "1.0")
        self._assert_version_order(u"1.0-alpha-1", "1.0-alpha-2")
        self._assert_version_order(u"1.0-alpha-1", "1.0-beta-1")
        self._assert_version_order(u"1.0-beta-1", "1.0-SNAPSHOT")
        self._assert_version_order(u"1.0-SNAPSHOT", "1.0")
        self._assert_version_order(u"1.0-alpha-1-SNAPSHOT", "1.0-alpha-1")
        self._assert_version_order(u"1.0", "1.0-1")
        self._assert_version_order(u"1.0-1", "1.0-2")
        self._assert_version_order(u"1.0.0", "1.0-1")
        self._assert_version_order(u"2.0-1", "2.0.1")
        self._assert_version_order(u"2.0.1-klm", "2.0.1-lmn")
        self._assert_version_order(u"2.0.1", "2.0.1-xyz")
        self._assert_version_order(u"2.0.1", "2.0.1-123")
        self._assert_version_order(u"2.0.1-xyz", "2.0.1-123")

        self._assert_version_order("1", u"2")
        self._assert_version_order("1.5", u"2")
        self._assert_version_order("1", u"2.5")
        self._assert_version_order("1.0", u"1.1")
        self._assert_version_order("1.1", u"1.2")
        self._assert_version_order("1.0.0", u"1.1")
        self._assert_version_order("1.0.1", u"1.1")
        self._assert_version_order("1.1", u"1.2.0")
        self._assert_version_order("1.0-alpha-1", u"1.0")
        self._assert_version_order("1.0-alpha-1", u"1.0-alpha-2")
        self._assert_version_order("1.0-alpha-1", u"1.0-beta-1")
        self._assert_version_order("1.0-beta-1", u"1.0-SNAPSHOT")
        self._assert_version_order("1.0-SNAPSHOT", u"1.0")
        self._assert_version_order("1.0-alpha-1-SNAPSHOT", u"1.0-alpha-1")
        self._assert_version_order("1.0", u"1.0-1")
        self._assert_version_order("1.0-1", u"1.0-2")
        self._assert_version_order("1.0.0", u"1.0-1")
        self._assert_version_order("2.0-1", u"2.0.1")
        self._assert_version_order("2.0.1-klm", u"2.0.1-lmn")
        self._assert_version_order("2.0.1", u"2.0.1-xyz")
        self._assert_version_order("2.0.1", u"2.0.1-123")
        self._assert_version_order("2.0.1-xyz", u"2.0.1-123")

        self._assert_version_order(u"1", u"2")
        self._assert_version_order(u"1.5", u"2")
        self._assert_version_order(u"1", u"2.5")
        self._assert_version_order(u"1.0", u"1.1")
        self._assert_version_order(u"1.1", u"1.2")
        self._assert_version_order(u"1.0.0", u"1.1")
        self._assert_version_order(u"1.0.1", u"1.1")
        self._assert_version_order(u"1.1", u"1.2.0")
        self._assert_version_order(u"1.0-alpha-1", u"1.0")
        self._assert_version_order(u"1.0-alpha-1", u"1.0-alpha-2")
        self._assert_version_order(u"1.0-alpha-1", u"1.0-beta-1")
        self._assert_version_order(u"1.0-beta-1", u"1.0-SNAPSHOT")
        self._assert_version_order(u"1.0-SNAPSHOT", u"1.0")
        self._assert_version_order(u"1.0-alpha-1-SNAPSHOT", u"1.0-alpha-1")
        self._assert_version_order(u"1.0", u"1.0-1")
        self._assert_version_order(u"1.0-1", u"1.0-2")
        self._assert_version_order(u"1.0.0", u"1.0-1")
        self._assert_version_order(u"2.0-1", u"2.0.1")
        self._assert_version_order(u"2.0.1-klm", u"2.0.1-lmn")
        self._assert_version_order(u"2.0.1", u"2.0.1-xyz")
        self._assert_version_order(u"2.0.1", u"2.0.1-123")
        self._assert_version_order(u"2.0.1-xyz", u"2.0.1-123")

    def test_compare(self):
        assert 1 < Version("1.0")
        assert VersionRange("1.0") == Version("1.0")

    def test_compare_errors(self):
        v = Version("1.0")
        self.assertRaises(RuntimeError, v._compare, v, 1.0)
        self.assertRaises(RuntimeError, v._int_compare, v, 1.0)
        self.assertRaises(RuntimeError, v._list_compare, v, 1.0)
        self.assertRaises(RuntimeError, v._string_compare, v, 1.0)


class TestVersionRange(unittest.TestCase):
    def test_no_lower_limit(self):
        vr = VersionRange("(,1.0]")
        assert len(vr.restrictions) == 1
        assert vr.version is None

    def test_single_spec(self):
        vr = VersionRange("1.0")
        assert len(vr.restrictions) == 1
        assert str(vr.version) == "1.0"

    def test_inclusive_range(self):
        vr = VersionRange("[1.0]")
        assert len(vr.restrictions) == 1
        assert vr.version is None

        vr = VersionRange("[1.2,1.3]")
        assert len(vr.restrictions) == 1
        assert vr.version is None

    def test_exclusive_upper_bound(self):
        vr = VersionRange("[1.0,2.0)")
        assert len(vr.restrictions) == 1
        assert vr.version is None

        vr = VersionRange("[1.5,)")
        assert len(vr.restrictions) == 1
        assert vr.version is None

    def test_multiple_restrictions(self):
        vr = VersionRange("(,1.0],[1.2,)")
        assert len(vr.restrictions) == 2
        assert vr.version is None

    def test_snapshots(self):
        vr = VersionRange("[1.0,)")
        assert "1.0-SNAPSHOT" not in vr

        vr = VersionRange("[1.0,1.1-SNAPSHOT]")
        assert "1.1-SNAPSHOT" in vr

        vr = VersionRange("[1.0,1.2]")
        assert "1.0-SNAPSHOT" not in vr
        assert "1.1-SNAPSHOT" in vr
        assert "1.2-SNAPSHOT" in vr
        assert "1.3-SNAPSHOT" not in vr

        vr = VersionRange("[1.0,1.2-SNAPSHOT]")
        assert "1.1-SNAPSHOT" in vr
        assert "1.2-SNAPSHOT" in vr

        vr = VersionRange("[1.0-SNAPSHOT,1.2]")
        assert "1.0-SNAPSHOT" in vr
        assert "1.1-SNAPSHOT" in vr

        vr = VersionRange("1.0-SNAPSHOT")
        assert "1.0-SNAPSHOT" in vr

        vr = VersionRange("[0.1,1.0)")
        assert "0.1" in vr
        assert "0.1.0" in vr
        assert "0.1-SNAPSHOT" not in vr
        assert "0.1.0-SNAPSHOT" not in vr

    def test_long_version(self):
        vr = VersionRange("[5.0.9.0,5.0.10.0)")
        assert Version("5.0.9.0") in vr
        assert "5.0.9.0" in vr

    def test_contains(self):
        test_pairs = (("2.0.5", True), ("2.0.4", True), ("[2.0.5]", True),
                      ("[2.0.6,)", False), ("[2.0.6]", False),
                      ("2.0,2.1]", True), ("[2.0,2.0.3]", False),
                      ("[2.0,2.0.5]", True), ("[2.0,2.0.5)", False),)
        v = Version("2.0.5")
        for spec, expected in test_pairs:
            vr = VersionRange(spec)
            assert (v in vr) == expected
            assert ("2.0.5" in vr) == expected

        assert VersionRange("2.0.5") in \
            VersionRange("[2.0,3.0)")

        assert VersionRange("[2.0.5]") not in \
            VersionRange("[2.0,3.0)")

    def test_invalid_ranges(self):
        for spec in ("(1.0,2.0", "[1.0,1.2),1.3", "[1.0,1.2),(1.1,1.3]",
                     "[1.1,1.3),(1.0,1.2]", "(1.1,1.2],[1.0,1.1)",):
            self.assertRaises(
                VersionRangeParseError, VersionRange, spec)

    def test_compare(self):
        vr1 = VersionRange("(1.0,2.0]")

        assert vr1 == vr1
        assert vr1 == "(1.0,2.0]"
        assert 1 < vr1

    def test_str(self):
        for input in ("[1.0,2.0]", "1.0"):
            actual = str(VersionRange(input))
            assert input == actual, \
                "VersionRange(%s) == %s, wanted %s" % (input, actual, input)

    def test_fromversion(self):
        v = Version("1.0")
        vr = VersionRange.from_version(v)

        assert vr.version == v

    def test_match_versions(self):
        versions = [Version("0.1"), Version("1.0"), Version("1.1"),
                    Version("2.0"), Version("2.1")]
        vr = VersionRange("(1.0,2.0]")
        assert vr.match_version(versions) == "2.0"
        assert vr.match_version(versions[:3]) == "1.1"



#
# Copyright (c) SAS Institute Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""
Exceptions used by pymaven
"""


# Generic errors
class PymavenError(Exception):
    """Base class for all errors in PymavenError

    Do not raise directly, but make a subclass
    """
    _template = None

    def __init__(self, *args, **kwargs):
        if self._template is None:
            super(PymavenError, self).__init__(*args, **kwargs)
        else:
            super(PymavenError, self).__init__(
                self._template.format(*args, **kwargs))


# Maven repo errors
class RepositoryError(PymavenError):
    """Generic errors raised by maven repositories"""


class MissingPathError(RepositoryError):
    """Raised when a repository accesses a path that does not exist"""
    _template = "No such directory: {0}"


# Maven Client errors
class ClientError(PymavenError):
    """Generic errors raised by maven clients"""


class MissingArtifactError(ClientError):
    """Raised when a client tries to fetch an artifact that does not exist"""
    _template = "No artifact found matching '{0}'"


# Parser errors
class ParseError(PymavenError):
    """Generic error in parsing a format"""


class ArtifactParseError(ParseError):
    """Raised when an error is encountered parsing maven coordinate"""


class RestrictionParseError(ParseError):
    """Raised when an error is encountered parsing a restriction spec"""


class VersionRangeParseError(ParseError):
    """Raised when an error is encountered parsing a range spec"""



#
# Copyright (c) SAS Institute Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import os
import tempfile
import unittest

from six import StringIO
import requests

from pymaven import Artifact
from pymaven.client import HttpRepository
from pymaven.client import LocalRepository
from pymaven.client import MavenClient
from pymaven.client import Struct
from pymaven.errors import MissingArtifactError
from pymaven.errors import MissingPathError

try:
    from unittest import mock
except ImportError:
    import mock


class TestMavenClient(unittest.TestCase):
    def test_invalid_repo(self):
        self.assertRaises(ValueError, MavenClient, "foo://bar.com")

    @mock.patch("pymaven.client.LocalRepository")
    def test_find_artifacts(self, _LocalRepository):
        _repo1 = mock.Mock(spec=LocalRepository)
        _repo2 = mock.Mock(spec=LocalRepository)

        _LocalRepository.side_effect = [_repo1, _repo2]

        _repo1.get_versions.return_value = [Artifact("foo:bar:2.0"),
                                            Artifact("foo:bar:2.0-SNAPSHOT"),
                                            Artifact("foo:bar:1.0"),
                                            Artifact("foo:bar:1.0-SNAPSHOT"),
                                            ]

        _repo2.get_versions.return_value = [Artifact("foo:bar:3.0"),
                                            Artifact("foo:bar:3.0-SNAPSHOT"),
                                            Artifact("foo:bar:1.0"),
                                            Artifact("foo:bar:1.0-SNAPSHOT"),
                                            ]
        client = MavenClient("foobar", "foobaz")
        expected = [Artifact("foo:bar:3.0"),
                    Artifact("foo:bar:3.0-SNAPSHOT"),
                    Artifact("foo:bar:2.0"),
                    Artifact("foo:bar:2.0-SNAPSHOT"),
                    Artifact("foo:bar:1.0"),
                    Artifact("foo:bar:1.0-SNAPSHOT"),
                    ]
        actual = client.find_artifacts("foo:bar")
        assert expected == actual, "client.find_artifacts(%s)" % input

    @mock.patch("pymaven.client.LocalRepository")
    def test_get_artifact(self, _LocalRepository):
        _repo = mock.Mock(spec=LocalRepository)

        _repo.exists.side_effect = [True]
        _repo.open.return_value = StringIO("some data")

        _LocalRepository.return_value = _repo

        client = MavenClient("/maven")
        actual = client.get_artifact("foo:bar:2.0.0")
        assert "some data" == actual.contents.read()
        _repo.exists.assert_called_with("foo/bar/2.0.0/bar-2.0.0.jar")
        _repo.open.assert_called_with("foo/bar/2.0.0/bar-2.0.0.jar")

    @mock.patch("pymaven.client.LocalRepository")
    def test_get_artifact_missing(self, _LocalRepository):
        _repo = mock.Mock(spec=LocalRepository)

        _repo.exists.side_effect = [False]
        _repo.open.return_value = StringIO("some data")

        _LocalRepository.return_value = _repo

        client = MavenClient("/maven")
        self.assertRaises(MissingArtifactError, client.get_artifact,
                          "foo:bar:3.0")
        _repo.exists.assert_called_with("foo/bar/3.0/bar-3.0.jar")
        _repo.open.assert_not_called()

    @mock.patch("pymaven.client.LocalRepository")
    def test_get_artifact_version_range(self, _LocalRepository):
        _repo = mock.Mock(spec=LocalRepository)

        _repo.exists.side_effect = [True, False]
        _repo.open.return_value = StringIO("some data")

        _LocalRepository.return_value = _repo

        client = MavenClient("/maven")
        self.assertRaises(AssertionError, client.get_artifact,
                          "foo:bar:[1.0,2.0]")
        _repo.open.assert_not_called()


@mock.patch("pymaven.client.HttpRepository._request")
class TestHttpRespository(unittest.TestCase):
    def test_listdir(self, _request):
        res = mock.MagicMock(spec=Struct)
        res.__enter__.return_value = StringIO(SIMPLE_METADATA)
        _request.side_effect = [res, requests.exceptions.HTTPError]
        expected = ["1.0-SNAPSHOT",
                    "1.0",
                    "3.0-SNAPSHOT",
                    "2.0.0",
                    "1.1",
                    ]
        repo = HttpRepository("http://foo.com/repo")
        actual = repo.listdir("foo/bar")
        assert expected == actual
        self.assertRaises(MissingPathError, repo.listdir, "/baz")

    def test_get_versions(self, _request):
        res = mock.MagicMock(spec=Struct)
        res.__enter__.return_value = StringIO(SIMPLE_METADATA)
        _request.return_value = res

        repo = HttpRepository("http://foo.com/repo")
        for input, expected in (
                ("foo:bar", [Artifact("foo:bar:3.0-SNAPSHOT"),
                             Artifact("foo:bar:2.0.0"),
                             Artifact("foo:bar:1.1"),
                             Artifact("foo:bar:1.0"),
                             Artifact("foo:bar:1.0-SNAPSHOT"),
                             ]),
                ("foo:bar:1.0", [Artifact("foo:bar:1.0")]),
                ("foo:bar:[1.0]", [Artifact("foo:bar:1.0")]),
                ("foo:bar:[1.0,2.0)", [Artifact("foo:bar:1.1"),
                                       Artifact("foo:bar:1.0"),
                                       ]),
                ("foo:bar:[2.0,3.0)", [Artifact("foo:bar:3.0-SNAPSHOT"),
                                       Artifact("foo:bar:2.0.0"),
                                       ]),
                ):
            actual = repo.get_versions(input)
            assert expected == actual, "HttpRepository.get_versions(%s)" % input
            # reset res contents
            res.__enter__.return_value.seek(0)

    def test_open(self, _request):
        res = mock.MagicMock(spec=Struct)
        res.__enter__.return_value = StringIO(SIMPLE_METADATA)
        _request.side_effect = [res, requests.exceptions.HTTPError]

        repo = HttpRepository("http://foo.com/repo")
        with repo.open("maven-metadata.xml") as fh:
            assert SIMPLE_METADATA == fh.read()
        self.assertRaises(MissingPathError, repo.open, "some/path")


class TestLocalRepository(unittest.TestCase):
    @mock.patch("pymaven.client.os")
    def test_get_versions(self, _os):
        _os.listdir.return_value = ["1.0-SNAPSHOT",
                                    "2.0.0",
                                    "3.0-SNAPSHOT",
                                    "1.1",
                                    "1.0",
                                    ]
        repo = LocalRepository("/maven")

        for input, expected in (
                ("foo:bar", [Artifact("foo:bar:3.0-SNAPSHOT"),
                             Artifact("foo:bar:2.0.0"),
                             Artifact("foo:bar:1.1"),
                             Artifact("foo:bar:1.0"),
                             Artifact("foo:bar:1.0-SNAPSHOT"),
                             ]),
                ("foo:bar:1.0", [Artifact("foo:bar:1.0")]),
                ("foo:bar:[1.0]", [Artifact("foo:bar:1.0")]),
                ("foo:bar:[1.0,2.0)", [Artifact("foo:bar:1.1"),
                                       Artifact("foo:bar:1.0"),
                                       ]),
                ("foo:bar:[2.0,3.0)", [Artifact("foo:bar:3.0-SNAPSHOT"),
                                       Artifact("foo:bar:2.0.0"),
                                       ]),
                ):
            actual = repo.get_versions(input)
            assert expected == actual, \
                "LocalRepository.get_versions(%s)" % input

    def test_open(self):
        with tempfile.NamedTemporaryFile(delete=False) as tmp:
            tmp.write(b"the file\n")

        try:
            repo = LocalRepository(os.path.dirname(tmp.name))
            with repo.open(tmp.name) as fh:
                assert "the file\n" == fh.read()

            self.assertRaises(MissingPathError, repo.open, "/does/not/exist")
        finally:
            # clean up temporary file
            os.remove(tmp.name)


SIMPLE_METADATA = """\
<?xml version="1.0" encoding="UTF-8"?>
<metadata>
  <groupId>foo</groupId>
  <artifactId>bar</artifactId>
  <version>3.0</version>
  <versioning>
    <latest>3.0</latest>
    <release>3.0</release>
    <versions>
      <version>1.0-SNAPSHOT</version>
      <version>1.0</version>
      <version>3.0-SNAPSHOT</version>
      <version>2.0.0</version>
      <version>1.1</version>
    </versions>
    <lastUpdated>20150521051651</lastUpdated>
  </versioning>
</metadata>
"""

FOO_BAR_3_0_POM = """\
<project xmlns="http://maven.apache.org/POM/4.0.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://maven.apache.org/POM/4.0.0 http://maven.apache.org/xsd/maven-4.0.0.xsd">
    <modelVersion>4.0.0</modelVersion>
    <groupId>foo</groupId>
    <artifactId>bar</artifactId>
    <version>2.0.0</version>
</project>
"""



#
# Copyright (c) SAS Institute Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from functools import wraps
from io import IOBase
from io import open
import posixpath

from six.moves.urllib.parse import urlsplit
from six.moves.urllib.parse import urlunsplit
import requests


def cmp(x, y):
    """
    Replacement for built-in funciton cmp that was removed in Python 3

    Compare the two objects x and y and return an integer according to
    the outcome. The return value is negative if x < y, zero if x == y
    and strictly positive if x > y.
    """
    if x is None and y is None:
        return 0
    elif x is None:
        return -1
    elif y is None:
        return 1
    return (x > y) - (x < y)


def memoize(name):
    def wrap(func):
        @wraps(func)
        def wrapper(slf, *args, **kwargs):
            if getattr(slf, name) is None:
                setattr(slf, name, func(slf, *args, **kwargs))
            return getattr(slf, name)
        return wrapper
    return wrap


def pad(seq, target_length, padding=None):
    """Extend the sequence *seq* with *padding* (default: None) if the length
    of *seq* is less than *target_length*. Modifies *seq* in place.

    >>> pad([], 5, 1)
    [1, 1, 1, 1, 1]
    >>> pad([1, 2, 3], 7)
    [1, 2, 3, None, None, None, None]
    >>> pad([1, 2, 3], 2)
    [1, 2, 3]

    :param list seq: list to padding
    :param int target_length: length to pad *seq* TestCommand
    :param padding: value to pad *seq* with
    :return: *seq* with appropriate padding
    """
    length = len(seq)
    if length < target_length:
        seq.extend([padding] * (target_length - length))
    return seq


def parse_source(source):
    """Parse ``source`` and return a file-like object"""
    if isinstance(source, IOBase):
        return source
    source_t = urlsplit(source)
    if source_t.scheme == "http" or source_t.scheme == "https":
        resp = requests.get(source, stream=True)
        resp.raise_for_status()
        return resp.raw
    if source_t.scheme == "file" or not source_t.scheme:
        return open(source_t.path, encoding="utf-8")
    raise TypeError("source must be a file name/path, file-like object, or URL")


def _first_of_each(*args, **kwargs):
    default = kwargs.get("default")
    return [next((s for s in sequence if s), default) for sequence in args]


def urljoin(*parts):
    """Join urls in a less surprising manner than urlparse.urljoin

    Inspired by this stackoverflow question:
    http://codereview.stackexchange.com/questions/13027/joining-url-path-components-intelligently

    >>> urljoin('http://foo.com', 'https://foo.com')
    'http://foo.com/'
    >>> urljoin('http://foo.com', 'bar')
    'http://foo.com/bar'
    >>> urljoin('http://foo.com/bar', 'baz')
    'http://foo.com/bar/baz'
    >>> urljoin('http://foo.com/bar', 'baz/spam')
    'http://foo.com/bar/baz/spam'
    >>> urljoin('http://foo.com', '/bar')
    'http://foo.com/bar'
    >>> urljoin('http://foo.com', 'bar', '?param=1&other=2')
    'http://foo.com/bar?param=1&other=2'
    >>> urljoin('http://foo.com', 'bar', '?param=1&other=2', '?not=3')
    'http://foo.com/bar?param=1&other=2'
    >>> urljoin('http://foo.com', 'bar', '#anchor', '#not')
    'http://foo.com/bar#anchor'
    >>> urljoin('http://foo.com', 'bar', '?param=1&other=2', '#anchor')
    'http://foo.com/bar?param=1&other=2#anchor'
    >>> urljoin('http://foo.com', 'bar', '?param=1&other=2', '#anchor',
    ...         '?not=3')
    'http://foo.com/bar?param=1&other=2#anchor'
    >>> urljoin('http://foo.com', 'bar', '?param=1&other=2', '#anchor', '#not')
    'http://foo.com/bar?param=1&other=2#anchor'
    >>> urljoin('http://foo.com', 'bar', '?param=1&other=2', '#anchor', '#not',
    ...         '?not=3')
    'http://foo.com/bar?param=1&other=2#anchor'
    >>> urljoin('#anchor', '?param=1&other=2', 'bar', 'http://foo.com')
    'http://foo.com/bar?param=1&other=2#anchor'
    """
    schemes, netlocs, paths, queries, fragments = \
        zip(*[urlsplit(part) for part in parts])
    scheme, netloc, query, fragment = _first_of_each(schemes, netlocs, queries,
                                                     fragments)
    path = posixpath.normpath(posixpath.join(*paths) or '/')
    return urlunsplit((scheme, netloc, path, query, fragment))



#
# Copyright (c) SAS Institute Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import getpass
import hashlib
import json
import logging
import os
import posixpath
import tempfile

from six.moves.urllib.parse import urlparse
import requests
import six

from . import utils
from .artifact import Artifact
from .errors import MissingArtifactError
from .errors import MissingPathError
from .pom import Pom
from .versioning import VersionRange

try:
    from xml.etree import cElementTree as ElementTree
except ImportError:
    from xml.etree import ElementTree

log = logging.getLogger(__name__)


class Struct(object):
    """ Simple object to mimic a requests.Response object
    """
    def __init__(self):
        self.status_code = None
        self.content = None
        self._json = None

    def __enter__(self):
        self._content = open(self.content, 'rb')
        return self._content

    def __exit__(self, exc_type, exc_value, traceback):
        self._content.close()

    @property
    @utils.memoize("_json")
    def json(self):
        with self as fh:
            return json.load(fh)

    def iter_content(self, size=None):
        with self as fh:
            for chunk in fh.read(size):
                yield chunk


class Cache(object):
    """ Local http cache
    """
    def __init__(self, cacheDir=None):
        if cacheDir is None:
            cacheDir = tempfile.mkdtemp(prefix=getpass.getuser())
        if not os.path.exists(cacheDir):
            os.makedirs(cacheDir, mode=0o700)
        self.cacheDir = cacheDir

    def _gen_key(self, method, uri, query_params):
        key = method + " " + uri
        if query_params:
            key += "?" + "&".join(
                ("%s=%s" % kv for kv in query_params.iteritems()))

        return key

    def _gen_hash(self, key):
        h = hashlib.sha1(key).hexdigest()
        return h

    def _gen_paths(self, hash):
        dhash = "%s.data" % hash
        hpath = os.path.join(self.cacheDir, hash)
        dhpath = os.path.join(self.cacheDir, dhash)
        return hpath, dhpath

    def cache(self, res, method, uri, query_params=None):
        """Access the cache for a request response

        :param str method: HTTP method
        :param str uri: location to requests
        :param dict query_params: query parameters
        :param requests.Response res: a request response to cache
        :return: the cached response
        """
        if query_params is None:
            query_params = {}

        key = self._gen_key(method, uri, query_params)
        h = self._gen_hash(key)
        hpath, dhpath = self._gen_paths(h)

        log.debug("Caching response %s with key %s", key, h)
        with open(hpath, "wb") as fh:
            for chunk in res.iter_content(1024):
                fh.write(chunk)
        with open(dhpath, "wb") as fh:
            json.dump({
                "status_code": res.status_code,
                "reason": res.reason,
                "method": method,
                "uri": uri,
                "param": query_params,
            }, fh)

        res = self._get(hpath, dhpath)
        return res

    def get(self, method, uri, query_params=None):
        if query_params is None:
            query_params = {}

        key = self._gen_key(method, uri, query_params)
        h = self._gen_hash(key)
        res = self._get(*self._gen_paths(h))
        if res is not None:
            log.debug("hit %s with key %s", key, h)
        else:
            log.debug("miss %s with key %s", key, h)
        return res

    def _get(self, hpath, dhpath):
        if os.path.exists(hpath) and os.path.exists(dhpath):
            data = json.load(open(dhpath))
            data["content"] = hpath
            res = Struct()
            for k, v in data.iteritems():
                setattr(res, k, v)
            return res


class MavenClient(object):
    """ Client for talking to a maven repository
    """
    def __init__(self, *urls):
        if isinstance(urls, six.string_types):
            urls = [urls]
        self._repos = []
        for url in urls:
            url = urlparse(url)
            if not url.scheme or url.scheme == "file":
                self._repos.append(LocalRepository(url.path))
            elif url.scheme.startswith("http"):
                self._repos.append(HttpRepository(url.geturl()))
            else:
                msg = "Unknown scheme: %s"
                log.error(msg, url)
                raise ValueError(msg % url.geturl())

    def find_artifacts(self, coordinate):
        """Find all artifacts matching the coordinate

        :param str coordinate: maven coordinate
        :retrun: list of Artifacts
        """
        artifacts = set([])
        for repo in self._repos:
            artifacts.update(set(repo.get_versions(coordinate)))
        return sorted(artifacts, reverse=True)

    def get_metadata(self, coordinate):
        """Return the metadata associated with the coordinates

        :param str coordinate: maven coordinate
        :raises: :py:exc:`pymaven.errors.MissingArtifactError`
        :return: the metadata requested
        :rtype: :py:class:`pymaven.pom.Pom`
        """
        query = Artifact(coordinate)
        assert query.version.version is not None, \
            "Cannot get metadata for version range"

        if query.type != "pom":
            query.type = "pom"

        for repo in self._repos:
            if repo.exists(query.path):
                return Pom(coordinate, self)
        else:
            raise MissingArtifactError(coordinate)

    def get_artifact(self, coordinate):
        """Return the actual artifact specified by the coordinate

        :param str coordinate: maven coordinate
        :raises: :py:exc:`pymaven.errors.MissingArtifactError`
        :return: the artifact requested
        :rtype: :py:class:`pymaven.Artifact`
        """
        query = Artifact(coordinate)
        assert query.version.version is not None, \
            "Cannot get artifact for version range"

        for repo in self._repos:
            if repo.exists(query.path):
                query.contents = repo.open(query.path)
                return query
        else:
            raise MissingArtifactError(coordinate)


class AbstractRepository(object):
    """
    Abstract class, this defines the interface that all repositories implement
    """
    def __init__(self, url):
        self._url = url

    def get_versions(self, coordinate):
        query = Artifact(coordinate)
        if self.exists(query.path):
            if query.version and query.version.version:
                return [Artifact(coordinate)]

            version_range = query.version
            if version_range is None:
                version_range = VersionRange("[,)")

            # base coordinate for all return values is everything up to the
            # version of the query
            base_coordinate = "%s:%s" % (query.group_id, query.artifact_id)
            if query.classifier:
                base_coordinate += "%s:%s" % (query.type, query.classifier)
            elif query.type != "jar":
                base_coordinate += ":%s" % query.type

            return sorted([
                Artifact(':'.join([base_coordinate, version]))
                for version in self.listdir(query.path)
                if version in version_range], reverse=True)

    def exists(self, path):
        """Return ``True`` if *path* exists in the repository, ``False``
        otherwise
        """
        return self._exists(path)

    def listdir(self, path):
        """List contents of *path*

        :raises: MissingPathError if *path* does not exist
        """
        try:
            return self._listdir(path)
        except (OSError, requests.exceptions.HTTPError):
            raise MissingPathError(path)

    def open(self, path):
        try:
            return self._open(path)
        except (IOError, requests.exceptions.HTTPError):
            raise MissingPathError("No such file: %s" % path)


class HttpRepository(AbstractRepository):
    """ Access a maven repository via http
    """
    def __init__(self, url, username=None, password=None):
        super(HttpRepository, self).__init__(url)
        self._cache = Cache()

    def _get(self, uri, **kwargs):
        res = self._request("GET", uri, **kwargs)
        return res

    def _head(self, uri, **kwargs):
        res = self._request("HEAD", uri, **kwargs)
        return res

    def _request(self, method, uri, json=False, **kwargs):
        url = utils.urljoin(self._url, uri)
        res = self._cache.get(method, uri, kwargs.get("params"))
        if not res:
            log.debug("requesting %s %s", method, url)
            res = requests.request(method, url, **kwargs)
            res = self._cache.cache(res, method, uri, kwargs.get("params"))

        if res.status_code != requests.codes.ok:
            raise requests.HTTPError(res.reason)

        if json:
            return res.json()
        return res

    def _exists(self, path):
        try:
            self._head(path)
        except requests.exceptions.HTTPError:
            return False
        return True

    def _listdir(self, path):
        uri = posixpath.join(path, "maven-metadata.xml")
        res = self._get(uri)
        with res as fh:
            metadata = ElementTree.parse(fh)

        return [elem.text.strip()
                for elem in metadata.findall("versioning/versions/version")
                if elem is not None]

    def _open(self, path):
        res = self._get(path, stream=True)
        return res


class LocalRepository(AbstractRepository):
    """A local disk-based repository
    """
    def _join(self, *args):
        return os.path.join(self._url, *args)

    def _exists(self, path):
        return os.path.exists(self._join(path))

    def _listdir(self, path):
        return os.listdir(self._join(path))

    def _open(self, path):
        """Open *path* as a file-like object

        The caller is responsible for calling close on the object
        """
        return open(self._join(path))



#
# Copyright (c) SAS Institute Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""
The artifact module provides objects and functions for working with artifacts
in a maven repository
"""

import functools
import re
import sys

import six

from .errors import ArtifactParseError
from .versioning import VersionRange

if sys.version_info > (2,):
    from .utils import cmp

MAVEN_COORDINATE_RE = re.compile(
    r'(?P<group_id>[^:]+)'
    r':(?P<artifact_id>[^:]+)'
    r'(:(?P<type>[^:]+)(:(?P<classifier>[^:]+))?)?'
    r':(?P<version>[^:])'
    )


@functools.total_ordering
class Artifact(object):
    """Represents an artifact within a maven repository."""

    __slots__ = ("group_id", "artifact_id", "version", "type", "classifier",
                 "contents")

    def __init__(self, coordinate):
        self.version = None
        self.type = "jar"
        self.classifier = None
        self.contents = None

        parts = coordinate.split(':')
        length = len(parts)
        if length < 2 or length > 5:
            raise ArtifactParseError(
                "Too many items in coordinate: '%s'" % coordinate)

        self.group_id, self.artifact_id = parts[:2]
        if length == 3:
            self.version = parts[2]
        elif length == 4:
            self.type = parts[2]
            self.version = parts[3]
        elif length == 5:
            self.type = parts[2]
            self.classifier = parts[3]
            self.version = parts[4]

        if self.version:
            self.version = VersionRange(self.version)

    def __cmp__(self, other):
        if self is other:
            return 0
        if not isinstance(other, Artifact):
            if isinstance(other, six.string_types):
                try:
                    return cmp(self, Artifact(other))
                except ArtifactParseError:
                    pass
            return 1
        result = cmp(self.group_id, other.group_id)
        if result == 0:
            result = cmp(self.artifact_id, other.artifact_id)
            if result == 0:
                result = cmp(self.type, other.type)
                if result == 0:
                    if self.classifier is None:
                        if other.classifier is not None:
                            result = 1
                    else:
                        if other.classifier is None:
                            result = -1
                        else:
                            result = cmp(self.classifier, other.classifier)
                    if result == 0:
                        result = cmp(self.version.version,
                                     other.version.version)
        return result

    def __eq__(self, other):
        return self.__cmp__(other) == 0

    def __lt__(self, other):
        return self.__cmp__(other) < 0

    def __ne__(self, other):
        return self.__cmp__(other) != 0

    def __hash__(self):
        return hash((self.group_id, self.artifact_id, self.version, self.type,
                     self.classifier))

    def __str__(self):
        s = ':'.join((self.group_id, self.artifact_id))
        if self.version:
            s += ':' + self.type
            if self.classifier:
                s += ':' + self.classifier
            s += ':' + str(self.version.version if self.version.version
                           else self.vserion)
        return s

    def __repr__(self):
        return "<pymaven.Artifact(%r)" % self.coordinate

    @property
    def coordinate(self):
        coordinate = "%s:%s" % (self.group_id, self.artifact_id)
        if self.type != "jar":
            coordinate += ":%s" % self.type
        if self.classifier is not None:
            coordinate += ":%s" % self.classifier
        if self.version is not None:
            coordinate += ":%s" % self.version
        return coordinate

    @property
    def path(self):
        path = "%s/%s" % (self.group_id.replace('.', '/'), self.artifact_id)

        if self.version and self.version.version:
            version = self.version.version
            path += "/%s/%s-%s" % (version, self.artifact_id, version)
            if self.classifier:
                path += "-%s" % self.classifier
            path += ".%s" % self.type
        return path



"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""
# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path


here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='python-maven',

    # https://packaging.python.org/en/latest/single_source_version.html
    version='0.0.3',

    description='Pymaven is a Python library for interfacing with the maven build system.',
    long_description=long_description,
    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Plugins',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Software Development'
    ],

    keywords='maven',
    packages=find_packages(exclude=['contrib', 'docs', 'test*']),

    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        'requests',
        'six',
        'lxml'
    ],
    extras_require={},
    package_data={},

    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    data_files=[],

    # Test suite (required for Py2)
    test_suite="tests",
)



#
# Copyright (c) SAS Institute Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import unittest

from six import BytesIO
import six

from pymaven import Artifact
from pymaven import VersionRange as VR
from pymaven.client import MavenClient
from pymaven.client import Struct
from pymaven.pom import Pom

try:
    from unittest import mock
except ImportError:
    import mock


class TestPom(unittest.TestCase):

    def _mock_client(self, *args):
        client = mock.MagicMock(spec=MavenClient)
        side_effect = []
        for arg in args:
            a = mock.MagicMock(spec=Artifact)
            a.contents = mock.MagicMock(spec=Struct)
            if isinstance(arg, six.string_types):
                arg = arg.encode("utf-8")
            a.contents.__enter__.return_value = BytesIO(arg)
            side_effect.append(a)
        client.get_artifact.side_effect = side_effect
        return client

    def test_parent(self):
        """Test pom parent processing"""
        client = self._mock_client(FOO_PARENT_1_POM)

        pom = Pom.fromstring("foo:bar:1", FOO_BAR_1_POM, client)
        assert pom.parent.group_id == "foo"
        assert pom.parent.artifact_id == "parent"
        assert pom.parent.version == "1"
        assert pom.parent.properties["groupId"] == "foo"
        assert pom.parent.properties["artifactId"] == "parent"
        assert pom.parent.properties["version"] == "1"
        assert pom.parent.properties["project.groupId"] == "foo"
        assert pom.parent.properties["project.artifactId"] == "parent"
        assert pom.parent.properties["project.version"] == "1"
        assert pom.parent.properties["pom.groupId"] == "foo"
        assert pom.parent.properties["pom.artifactId"] == "parent"
        assert pom.parent.properties["pom.version"] == "1"
        assert pom.properties["parent.groupId"] == "foo"
        assert pom.properties["parent.artifactId"] == "parent"
        assert pom.properties["parent.version"] == "1"
        client.get_artifact.assert_called_with("foo:parent:pom:1")

    def test_replace_properties(self):
        """Test Pom._replace_properties"""
        client = self._mock_client(FOO_BAR_1_POM, FOO_PARENT_1_POM)
        pom = Pom("foo:bar:1", client)
        properties = {
            "prop1": "\na string\n",
            "prop2": "${prop1}",
        }
        assert "prop1" == pom._replace_properties("prop1", properties)
        assert "a string" == pom._replace_properties("${prop1}", properties)
        assert "a string" == pom._replace_properties("${prop2}", properties)
        assert "baz version string" == pom._replace_properties("${bazChild}")
        assert "baz version string" == pom._replace_properties("${bazVersion}")
        assert "${unmatched}" == pom._replace_properties("${unmatched}")
        assert "${parentProp}" == \
            pom.parent._replace_properties("${parentProp}")
        assert "resolve" == pom._replace_properties("${resolveProp}")
        assert "resolve" == pom._replace_properties("${parentProp}")

    def test_find_relocations(self):
        """Test Pom._find_relocations()"""
        for args, coordinate in (
                ((RELOCATION_1, FOO_BAR_1_POM), "foo.org:bar:pom:1"),
                ((RELOCATION_2, FOO_BAR_1_POM), "foo:baz:pom:1"),
                ((RELOCATION_3, FOO_BAR_1_POM), "foo:bar:pom:alpha"),
                ((RELOCATION_4, FOO_BAR_1_POM), "foo.org:baz:pom:alpha"),
                ):
            client = self._mock_client(*args)
            pom = Pom(coordinate, client)
            assert len(pom.dependencies["relocation"]) == 1
            relocations = list(pom.dependencies["relocation"])
            assert relocations[0] == (("foo", "bar", VR("1")), True)

    def test_find_prereqs(self):
        """Test Pom._find_prerequisites()"""
        client = self._mock_client(PREREQUISITES_1)
        pom = Pom("foo:bar:1", client)

        assert "prereq1" == pom.properties["prerequisites.one"]
        assert "prereq1" == pom.properties["project.prerequisites.one"]
        assert "prereq2" == pom.properties["prerequisites.two"]
        assert "prereq2" == pom.properties["project.prerequisites.two"]

    def test_find_import_deps(self):
        """Test Pom._find_import_deps()"""
        client = self._mock_client(IMPORT_DEPS_1, FOO_PARENT_1_POM,
                                   FOO_PARENT_1_POM)
        pom = Pom("foo:bar:1", client)

        import_deps = list(pom._find_import_deps()["import"])
        assert import_deps[0] == (("foo", "parent", "1"), True)

    def test_dependency_management(self):
        """Test Pom.dependency_management"""
        client = self._mock_client(COM_TEST_USE, COM_TEST_BOM, COM_TEST_BOM2)
        pom = Pom("com.test:use:1", client)

        dep_mgmt = pom.dependency_management
        assert ("1.0.0", "import", False) == dep_mgmt[("com.test", "bom")]
        assert ("2.0.0", "import", False) == dep_mgmt[("com.test", "bom2")]
        assert ("2.0.0", None, True) == dep_mgmt[("com.test", "project1")]
        assert ("1.0.0", None, False) == dep_mgmt[("com.test", "project2")]

    def test_deps(self):
        client = self._mock_client(COM_TEST_DEP, COM_TEST_PROJECT1,
                                   COM_TEST_PROJECT2)
        pom = Pom("com.test:dep:1.0.0", client)

        compile_deps = list(pom.dependencies["compile"])
        runtime_deps = list(pom.dependencies["runtime"])

        assert len(compile_deps) == 1
        assert compile_deps[0] == (("com.test", "project1", "1.0.0"), True)

        assert len(runtime_deps) == 1
        assert runtime_deps[0] == (("com.test", "project2", "1.0.0"), True)

    def test_dependency_version_range(self):
        client = self._mock_client(COM_TEST_PROJECT3, COM_TEST_PROJECT2)
        client.find_artifacts.return_value = [
            Artifact("com.test:project2:2.0.0"),
            Artifact("com.test:project2:1.0.0"),
            ]
        pom = Pom("com.test:project3:1.0.0", client)

        deps = list(pom.dependencies["compile"])
        assert len(pom.dependencies) == 1
        assert len(deps) == 1
        assert deps[0] == (("com.test", "project2", "[1.0,2.0)"), True)

        client = self._mock_client(COM_TEST_PROJECT4)
        client.find_artifacts.return_value = [
            Artifact("com.test:project2:2.0.0-SNAPSHOT"),
            Artifact("com.test:project2:1.0.0"),
            ]
        pom = Pom("com.test:project4:1.0.0", client)

        deps = list(pom.dependencies["compile"])
        assert len(pom.dependencies) == 1
        assert len(deps) == 1
        assert deps[0] == (("com.test", "project2", "release"), True)

        client = self._mock_client(
            COM_TEST_PROJECT4.replace("version>release", "version>latest"))
        client.find_artifacts.return_value = [
            Artifact("com.test:project2:2.0.0-SNAPSHOT"),
            Artifact("com.test:project2:1.0.0"),
            ]
        pom = Pom("com.test:project4:1.0.0", client)

        deps = list(pom.dependencies["compile"])
        assert len(pom.dependencies) == 1
        assert len(deps) == 1
        assert deps[0] == (("com.test", "project2", "latest"), True)

        client = self._mock_client(
            COM_TEST_PROJECT4.replace("version>release",
                                      "version>latest.release"))
        client.find_artifacts.return_value = [
            Artifact("com.test:project2:2.0.0-SNAPSHOT"),
            Artifact("com.test:project2:1.0.0"),
            ]
        pom = Pom("com.test:project4:1.0.0", client)

        deps = list(pom.dependencies["compile"])
        assert len(pom.dependencies) == 1
        assert len(deps) == 1
        assert deps[0] == (("com.test", "project2", "latest.release"), True)

        client = self._mock_client(
            COM_TEST_PROJECT4.replace("version>release",
                                      "version>latest.integration"))
        client.find_artifacts.return_value = [
            Artifact("com.test:project2:2.0.0-SNAPSHOT"),
            Artifact("com.test:project2:1.0.0"),
            ]
        pom = Pom("com.test:project4:1.0.0", client)

        deps = list(pom.dependencies["compile"])
        assert len(pom.dependencies) == 1
        assert len(deps) == 1
        assert deps[0] == (("com.test", "project2", "latest.integration"), True)

    def test_profiles(self):
        client = self._mock_client(COM_TEST_PROFILE_1, COM_TEST_PROJECT1)
        pom = Pom("com.test:profile:1.0.0", client)

        compile_deps = list(pom.dependencies["compile"])
        assert "true" == pom.properties["active_profile"]
        assert compile_deps[0] == (("com.test", "project1", VR("1.0.0")), True)

        for input, expected in (
                ("[1.5,", "true"),
                ("![1.5,", "false"),
                ("!1.5", "true"),
                ("1.5", "false"),
                ("![1.5,1.7]", "true"),
                ("[1.5,1.7]", "false"),
                ("1.8", "true"),
                ("!1.8", "false"),
                ("[1.8,)", "true"),
                ("![1.8,)", "false"),
                ("![1.5,1.8)", "true"),
                ("[1.5,1.8)", "false"),
                ("![,1.8)", "true"),
                ("[,1.8)", "false"),
                ("[,1.8]", "true"),
                ("![,1.8]", "false"),
                ):
            client = self._mock_client(
                COM_TEST_PROFILE_2.replace("@JDK@", input))
            pom = Pom("com.test:profile:1.0.0", client)
            actual = pom.properties["default_profile"]
            assert expected == actual, \
                "%s: Wanted %s, got %s" % (input, expected, actual)

COM_TEST_PROFILE_1 = """\
<project xmlns="http://maven.apache.org/POM/4.0.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
    xsi:schemaLocation="http://maven.apache.org/POM/4.0.0 http://maven.apache.org/xsd/maven-4.0.0.xsd">
    <modelVersion>4.0.0</modelVersion>
    <groupId>com.test</groupId>
    <artifactId>profile</artifactId>
    <version>1.0.0</version>
    <profiles>
        <profile>
            <activation>
                <activeByDefault>true</activeByDefault>
            </activation>
            <properties>
                <active_profile>true</active_profile>
            </properties>
            <dependencies>
                <dependency>
                    <groupId>com.test</groupId>
                    <artifactId>project1</artifactId>
                </dependency>
            </dependencies>
            <dependencyManagement>
                <dependencies>
                    <dependency>
                        <groupId>com.test</groupId>
                        <artifactId>project1</artifactId>
                        <version>1.0.0</version>
                    </dependency>
                </dependencies>
            </dependencyManagement>
        </profile>
    </profiles>
</project>
"""

COM_TEST_DEP = """\
<project xmlns="http://maven.apache.org/POM/4.0.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
    xsi:schemaLocation="http://maven.apache.org/POM/4.0.0 http://maven.apache.org/xsd/maven-4.0.0.xsd">
    <modelVersion>4.0.0</modelVersion>
    <groupId>com.test</groupId>
    <artifactId>project1</artifactId>
    <version>1.0.0</version>
    <dependencies>
        <dependency>
            <groupId>com.test</groupId>
            <artifactId>project1</artifactId>
            <version>1.0.0</version>
        </dependency>
        <dependency>
            <groupId>com.test</groupId>
            <artifactId>project2</artifactId>
            <version>1.0.0</version>
            <scope>runtime</scope>
        </dependency>
    </dependencies>
</project>
"""

COM_TEST_PROFILE_2 = """\
<project xmlns="http://maven.apache.org/POM/4.0.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
    xsi:schemaLocation="http://maven.apache.org/POM/4.0.0 http://maven.apache.org/xsd/maven-4.0.0.xsd">
    <modelVersion>4.0.0</modelVersion>
    <groupId>com.test</groupId>
    <artifactId>profile</artifactId>
    <version>1.0.0</version>
    <profiles>
        <profile>
            <activation>
                <activeByDefault>true</activeByDefault>
            </activation>
            <properties>
                <default_profile>false</default_profile>
            </properties>
        </profile>
        <profile>
            <activation>
                <jdk>@JDK@</jdk>
            </activation>
            <properties>
                <default_profile>true</default_profile>
            </properties>
        </profile>
    </profiles>
</project>
"""

COM_TEST_PROJECT1 = """\
<project xmlns="http://maven.apache.org/POM/4.0.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
    xsi:schemaLocation="http://maven.apache.org/POM/4.0.0 http://maven.apache.org/xsd/maven-4.0.0.xsd">
    <modelVersion>4.0.0</modelVersion>
    <groupId>com.test</groupId>
    <artifactId>project1</artifactId>
    <version>1.0.0</version>
</project>
"""

COM_TEST_PROJECT2 = """\
<project xmlns="http://maven.apache.org/POM/4.0.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
    xsi:schemaLocation="http://maven.apache.org/POM/4.0.0 http://maven.apache.org/xsd/maven-4.0.0.xsd">
    <modelVersion>4.0.0</modelVersion>
    <groupId>com.test</groupId>
    <artifactId>project2</artifactId>
    <version>1.0.0</version>
</project>
"""

COM_TEST_PROJECT3 = """\
<project xmlns="http://maven.apache.org/POM/4.0.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
    xsi:schemaLocation="http://maven.apache.org/POM/4.0.0 http://maven.apache.org/xsd/maven-4.0.0.xsd">
    <modelVersion>4.0.0</modelVersion>
    <groupId>com.test</groupId>
    <artifactId>project1</artifactId>
    <version>1.0.0</version>
    <dependencies>
        <dependency>
            <groupId>com.test</groupId>
            <artifactId>project2</artifactId>
            <version>[1.0,2.0)</version>
        </dependency>
    </dependencies>
</project>
"""

COM_TEST_PROJECT4 = """\
<project xmlns="http://maven.apache.org/POM/4.0.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
    xsi:schemaLocation="http://maven.apache.org/POM/4.0.0 http://maven.apache.org/xsd/maven-4.0.0.xsd">
    <modelVersion>4.0.0</modelVersion>
    <groupId>com.test</groupId>
    <artifactId>project1</artifactId>
    <version>1.0.0</version>
    <dependencies>
        <dependency>
            <groupId>com.test</groupId>
            <artifactId>project2</artifactId>
            <version>release</version>
        </dependency>
    </dependencies>
</project>
"""

COM_TEST_METADATA = """\
<?xml version="1.0" encoding="UTF-8"?>
<metadata>
    <groupId>com.test</groupId>
    <artifactId>project2</artifactId>
    <versioning>
        <latest>2.0.0</latest>
        <release>1.0.0</release>
        <versions>
            <version>2.0.0</version>
            <version>1.0.0</version>
        </versions>
    </versioning>
</metadata>
"""

COM_TEST_BOM2 = """\
<project xmlns="http://maven.apache.org/POM/4.0.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
    xsi:schemaLocation="http://maven.apache.org/POM/4.0.0 http://maven.apache.org/xsd/maven-4.0.0.xsd">
  <modelVersion>4.0.0</modelVersion>
  <groupId>com.test</groupId>
  <artifactId>bom2</artifactId>
  <version>2.0.0</version>
  <packaging>pom</packaging>
  <properties>
    <project1Version>2.0.0</project1Version>
  </properties>
  <dependencyManagement>
    <dependencies>
      <dependency>
        <groupId>com.test</groupId>
        <artifactId>project1</artifactId>
        <version>${project1Version}</version>
        <optional>true</optional>
      </dependency>
    </dependencies>
  </dependencyManagement>
  <modules>
    <module>parent</module>
  </modules>
</project>
"""

COM_TEST_BOM = """\
<project xmlns="http://maven.apache.org/POM/4.0.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
    xsi:schemaLocation="http://maven.apache.org/POM/4.0.0 http://maven.apache.org/xsd/maven-4.0.0.xsd">
  <modelVersion>4.0.0</modelVersion>
  <groupId>com.test</groupId>
  <artifactId>bom</artifactId>
  <version>1.0.0</version>
  <packaging>pom</packaging>
  <properties>
    <project1Version>1.0.0</project1Version>
    <project2Version>1.0.0</project2Version>
  </properties>
  <dependencyManagement>
    <dependencies>
      <dependency>
        <groupId>com.test</groupId>
        <artifactId>project1</artifactId>
        <version>${project1Version}</version>
      </dependency>
      <dependency>
        <groupId>com.test</groupId>
        <artifactId>project2</artifactId>
        <version>${project1Version}</version>
      </dependency>
    </dependencies>
  </dependencyManagement>
  <modules>
    <module>parent</module>
  </modules>
</project>
"""

COM_TEST_USE = """\
<project xmlns="http://maven.apache.org/POM/4.0.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
     xsi:schemaLocation="http://maven.apache.org/POM/4.0.0 http://maven.apache.org/xsd/maven-4.0.0.xsd">
  <modelVersion>4.0.0</modelVersion>
  <groupId>com.test</groupId>
  <artifactId>use</artifactId>
  <version>1.0.0</version>
  <packaging>jar</packaging>

  <dependencyManagement>
    <dependencies>
      <dependency>
        <groupId>com.test</groupId>
        <artifactId>bom</artifactId>
        <version>1.0.0</version>
        <type>pom</type>
        <scope>import</scope>
      </dependency>
      <dependency>
        <groupId>com.test</groupId>
        <artifactId>bom2</artifactId>
        <version>2.0.0</version>
        <type>pom</type>
        <scope>import</scope>
      </dependency>
    </dependencies>
  </dependencyManagement>
  <dependencies>
    <dependency>
      <groupId>com.test</groupId>
      <artifactId>project1</artifactId>
    </dependency>
    <dependency>
      <groupId>com.test</groupId>
      <artifactId>project2</artifactId>
      <optional>true</optional>
    </dependency>
  </dependencies>
</project>
"""

IMPORT_DEPS_1 = """\
<?xml version="1.0" encoding="UTF-8"?>
<project xmlns="http://maven.apache.org/POM/4.0.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://maven.apache.org/POM/4.0.0 http://maven.apache.org/xsd/maven-4.0.0.xsd">
    <modelVersion>4.0.0</modelVersion>
    <groupId>foo</groupId>
    <artifactId>bar</artifactId>
    <version>1</version>
    <dependencyManagement>
        <dependencies>
            <dependency>
                <groupId>foo</groupId>
                <artifactId>parent</artifactId>
                <version>1</version>
                <scope>import</scope>
            </dependency>
            <dependency>
                <groupId>foo</groupId>
                <artifactId>baz</artifactId>
                <version>1</version>
            </dependency>
            <dependency>
                <groupId>foo</groupId>
                <artifactId>spam</artifactId>
                <version>1</version>
                <scope>compile</scope>
            </dependency>
        </dependencies>
    </dependencyManagement>
</project>
"""

PREREQUISITES_1 = """\
<?xml version="1.0" encoding="UTF-8"?>
<project xmlns="http://maven.apache.org/POM/4.0.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://maven.apache.org/POM/4.0.0 http://maven.apache.org/xsd/maven-4.0.0.xsd">
    <modelVersion>4.0.0</modelVersion>
    <groupId>foo.org</groupId>
    <artifactId>bar</artifactId>
    <version>1</version>
    <prerequisites>
        <one>prereq1</one>
        <two>prereq2</two>
    </prerequisites>
</project>
"""

RELOCATION_1 = """\
<?xml version="1.0" encoding="UTF-8"?>
<project xmlns="http://maven.apache.org/POM/4.0.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://maven.apache.org/POM/4.0.0 http://maven.apache.org/xsd/maven-4.0.0.xsd">
    <modelVersion>4.0.0</modelVersion>
    <groupId>foo.org</groupId>
    <artifactId>bar</artifactId>
    <version>1</version>
    <distributionManagement>
        <relocation>
            <groupId>foo</groupId>
        </relocation>
    </distributionManagement>
</project>
"""

RELOCATION_2 = """\
<?xml version="1.0" encoding="UTF-8"?>
<project xmlns="http://maven.apache.org/POM/4.0.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://maven.apache.org/POM/4.0.0 http://maven.apache.org/xsd/maven-4.0.0.xsd">
    <modelVersion>4.0.0</modelVersion>
    <groupId>foo</groupId>
    <artifactId>baz</artifactId>
    <version>1</version>
    <distributionManagement>
        <relocation>
            <artifactId>bar</artifactId>
        </relocation>
    </distributionManagement>
</project>
"""

RELOCATION_3 = """\
<?xml version="1.0" encoding="UTF-8"?>
<project xmlns="http://maven.apache.org/POM/4.0.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://maven.apache.org/POM/4.0.0 http://maven.apache.org/xsd/maven-4.0.0.xsd">
    <modelVersion>4.0.0</modelVersion>
    <groupId>foo</groupId>
    <artifactId>bar</artifactId>
    <version>alpha</version>
    <distributionManagement>
        <relocation>
            <version>1</version>
        </relocation>
    </distributionManagement>
</project>
"""

RELOCATION_4 = """\
<?xml version="1.0" encoding="UTF-8"?>
<project xmlns="http://maven.apache.org/POM/4.0.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://maven.apache.org/POM/4.0.0 http://maven.apache.org/xsd/maven-4.0.0.xsd">
    <modelVersion>4.0.0</modelVersion>
    <groupId>foo.org</groupId>
    <artifactId>baz</artifactId>
    <version>alpha</version>
    <distributionManagement>
        <relocation>
            <groupId>foo</groupId>
            <artifactId>bar</artifactId>
            <version>1</version>
        </relocation>
    </distributionManagement>
</project>
"""

FOO_BAR_1_POM = """\
<?xml version="1.0" encoding="UTF-8"?>
<project xmlns="http://maven.apache.org/POM/4.0.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://maven.apache.org/POM/4.0.0 http://maven.apache.org/xsd/maven-4.0.0.xsd">
    <modelVersion>4.0.0</modelVersion>
    <parent>
        <groupId>foo</groupId>
        <artifactId>parent</artifactId>
        <version>1</version>
    </parent>
    <artifactId>bar</artifactId>
    <properties>
        <bazChild>${bazVersion}</bazChild>
        <resolveProp>resolve</resolveProp>
    </properties>
</project>
"""

FOO_PARENT_1_POM = """\
<?xml version="1.0" encoding="UTF-8"?>
<project xmlns="http://maven.apache.org/POM/4.0.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://maven.apache.org/POM/4.0.0 http://maven.apache.org/xsd/maven-4.0.0.xsd">
    <modelVersion>4.0.0</modelVersion>
    <groupId>foo</groupId>
    <artifactId>parent</artifactId>
    <version>1</version>
    <properties>
        <parentProp>${resolveProp}</parentProp>
        <bazVersion>baz version string
</bazVersion>
    </properties>
</project>
"""
